/*
 *------------------------------------------------------------------
 *
 * queue.c
 * Nov 98
 * 304-427A
 * Sorin Lerner (based on original implementation from
 * prof. Cooperstock)
 *
 * Queue routines for threads library. In all routines, we
 * assume that queue pointer is not NULL and the TD pointer
 * is not NULL.
 *
 *------------------------------------------------------------------
 */

#include <stdio.h>
#include <assert.h>
#include "threadimpl.h"
#include "queue.h"


/*
 * Create a queue. Returns NULL if out of memory.
 */

Queue* QueueCreate(int bOwn) {

  Queue* result;

  result = (Queue*)malloc(sizeof(Queue));

  if (result) {
    result->bOwn = bOwn;
    result->head = NULL;
    result->tail = NULL;
  }

  return result;

}


/*
 * Destroy the given queue.
 */

void QueueDestroy(Queue* queue) {

  assert(queue);

  if (queue->bOwn) {

    /* We own the data, so free the whole list */

    TD* td = queue->head;
    while (td) {
      TD* next = td->next;
      free(td);
      td = next;
    }

  }

  free(queue);

}

/*
 * Return true if empty.
 */

int QueueIsEmpty(Queue* queue) {
  assert(queue);
  return (NULL == queue->head);
}

/*
 * Enqueue the thread descriptor onto the queue (ie: add to the tail)
 */

void QueueEnqueue(Queue* queue, TD* td) {

  assert(queue);
  assert(td);

  td->next = NULL;
  td->inqueue = queue;

  if (queue->tail) 
    queue->tail->next = td;
  else
    queue->head = td;

  queue->tail = td;

}

/*
 * remove and return item from head of queue
 */

TD* QueueDequeue(Queue *queue) {

  TD* td = NULL;

  assert(queue);

  if (queue->head) {
    td = queue->head;
    td->inqueue = NULL;
    queue->head = td->next;
    if (NULL == queue->head)
      queue->tail = NULL;
  }

  return td;

}

/*
 * Insert td into queue, according to the priority field.
 */ 

void QueueInsert(Queue* queue,TD *td) {

  TD *prev, *curr;

  assert(queue);
  assert(td);

  td->inqueue = queue;

  prev = NULL;
  curr = queue->head;

  while (curr && td->priority >= curr->priority) {
    prev = curr;
    curr = curr->next;
  }

  td->next = curr;

  if (curr == NULL)
    queue->tail = td;

  if (prev == NULL)
    queue->head = td;
  else
    prev->next = td;

}

/*
 * Deletes the TD from the given queue. Returns the given TD if successfull,
 * and NULL if the TD was not found.
 */

TD* QueueDelete(Queue* queue,TD* td) {

  TD *prev, *curr;

  assert(queue);
  assert(td);

  prev = NULL;
  curr = queue->head;

  while (curr && curr != td) {
    prev = curr;
    curr = curr->next;
  }

  if (curr) {
    td->inqueue = NULL;

    if (curr == queue->tail)
      queue->tail = prev;

    if (prev)
      prev->next = curr->next;
    else
      queue->head = curr->next;
  }

  return curr;
}

